<?php
/**
 * Copyright (c) 2005-2008 GForge, LLC
 *
 * @version   $Id$
 * @author Marcelo Mottalli	marcelo@gforgegroup.com
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once('SystemUtils.php');
require_once('plugins/scm/lib/SCMRepository.php');

/**
 * Class that redirects requests to the gitweb interface
 */
class GitwebBrowser {
	private $repos;
	
	public function __construct(SCMRepository $repos) {
		$this->repos = $repos;
		$this->gitweb_install = SystemUtils::buildFilePath(GFConf::get('system.libDir'), '/plugins/scmgit/lib/external/gitweb/');
	}

	public function parseQuery($querystring) {
		// read the environment variables for gitweb from the query string
		$vars = array();
		parse_str($querystring, $vars);
		$this->buildEnvironment($querystring);
		
		// Locate gitweb.cgi
		$gitweb = SystemUtils::buildFilePath($this->gitweb_install, '/gitweb.cgi');
		if (!file_exists($gitweb)) {
			throw new Exception('gitweb must be installed to '.$gitweb_install);
		}
		
		// Build the command to execute
		$cmd = '';
		foreach ($this->environ as $var => $val) {
			$cmd .= $var.'='.escapeshellarg($val).' ';
		}
		$cmd .= $gitweb;
		$cmd .= ' 2>&1';		// redirect stderr to stdout
		
		// Execute gitweb
		$output = '';
		// We don't want the output to be exploded since it may be a binary result.
		// PHP has a nasty problem where it treats both \n and \r\n as new lines, and
		// it removes both of them, so we don't have a way to reconstruct binary files
		$result = SystemUtils::execute($cmd, $output, false);
		$output = explode("\n", $output);
		
		if ($result != 0) {		// error
			$this->isHtml = true;
			$this->content = implode("\n", $output);
			return;
		}

		// Parse the headers
		$eoh = false;
		$this->contentType = 'text/html';
		$this->headers = array();
		while (true) {		// delete the headers
			$line = trim(array_shift($output));
			
			// First empty line marks the end of the HTTP headers
			$eoh = empty($line);
			if ($eoh) {
				break;
			}

			if (preg_match('/Content-type: (.*)/i', $line, $matches)) {
				$this->contentType = $matches[1];
			}
			$this->headers[] = $line;

		};
		
		if (strpos($this->contentType, 'text/html') === false) {
			$this->isHtml = false;
			// No further parsing is needed
			$this->content = implode("\n", $output);
			return;
		}
		
		// Now we need to parse the HTML generated by gitweb in order to change some URLs
		$this->isHtml = true;
		$this->content = $this->parseOutput($output);
	}
	
	private function parseOutput($output) {
		$theme = Session::getSession()->getTheme();
		$newOutput = array();
		
		// Exclude everything outside <body>...</body>
		$body = array();
		$in_body = false;
		for ($i = 0; $i < count($output); $i++) {
			$line = trim($output[$i]);
			if ($line == '<body>') {
				$in_body = true;
				continue;
			} elseif ($line == '</body>') {
				break;
			}
			
			if ($in_body) $body[] = $line;
		}
		$output = $body;

		for ($i=0; $i < count($output); $i++) {
			$line = $output[$i];
			
			$keep_parsing_line = true;
			
			//TODO: check if the line needs to be parsed
			
			if (!$keep_parsing_line) {
				// Move to the next line, don't parse the links
				continue;
			}
			
			if (false) {
				// no need to parse lines
			} else {
				// Unfortunately we can't link to the tracker item since the log message always links to the
				// commit (we can't add a link inside another link)
				/*preg_match_all('/\\[#(\d+)\\]/', $line, $matches);
				if (count($matches) > 1) {
					// Found a tracker item ID like [#nnn]
					$tracker_item_ids = $matches[1];
					foreach ($tracker_item_ids as $tracker_item_id) {
						$link = GFUrl::getUrl()->projectUrl($this->repos->getProject(), 'TrackerItemEdit', 'tracker', '', array('tracker_item_id' => $tracker_item_id));
						$line = preg_replace('/\\[#'.$tracker_item_id.'\\]/', '<a href="'.$link.'">[#'.$tracker_item_id.']</a>', $line);
					}
				}*/
				$newOutput[] = $line;
			}
		}
		
		// If we output html and we found the mbstring extension, we
		// should try to encode the output of ViewVC in UTF-8
		$content = implode("\n", $newOutput);
		if (extension_loaded('mbstring')) {
			$encoding = mb_detect_encoding($content);
			if ($encoding != 'UTF-8') {
				$content = mb_convert_encoding($content, 'UTF-8', $encoding);
			}
		}
		
		return $content;		
	}

	private function buildEnvironment($querystring) {
		// Translate a GForge querystring to a gitweb querystring
		$vars = array();
		parse_str($querystring, $vars);
		
		// Check if it's a GForge querystring. Note that gitweb uses a different format for querystrings
		// so parse_str won't work as expected
		if (empty($vars) || isset($vars['action']) || isset($vars['path'])) {
			// Translate the GForge variables to a querystring gitweb will understand
			$querystring = 'p='.$this->repos->getRepositoryName();
			if (isset($vars['path'])) {
				$path = preg_replace('/^\\//', '', $vars['path']);		// remove trailing /
				$querystring .= ';f='.urlencode($path);
				
				if (isset($vars['r1']) && isset($vars['r2'])) {
					// Making a diff...
					$querystring .= ';a=blobdiff;hb='.$vars['r2'].';hpb='.$vars['r1'];
				} else {
					// Just viewing the history of a file
					$querystring .= ';a=history';
					if (isset($vars['newversion'])) {
						$querystring .= ';hb='.$vars['newversion'];				
					}
				}
			}
		} else {
			// $querystring already has the gitweb format -- do nothing
		}

		// Security check: the "p" (project) parameter MUST be set and MUST be equal to the project name
		if (!preg_match('/^p='.$this->repos->getRepositoryName().'/', $querystring)) {
			throw new PermissionDeniedException();
		}

		$this->environ['QUERY_STRING'] = $querystring;
		$this->environ['PROJECTROOT'] = dirname($this->repos->getRoot());
	}

	public function getContentType() {
		return $this->contentType;
	}
	
	public function isHtml() {
		return $this->isHtml;
	}
	
	public function getContents() {
		return $this->content;
	}
	
	public function getHTTPHeaders() {
		return $this->headers;
	}
	

}
?>
