<?php
/**
 * Copyright (c) 2005-2008 GForge, LLC
 *
 * @version   $Id: parse_svn_history.php 3226 2008-08-06 17:58:08Z bigdisk $
 * @author Marcelo Mottalli	marcelo@gforgegroup.com
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */
 

require_once("system/Cronjob.php");
require_once('core/CronHistory.php');
require_once('plugins/scmsvn/lib/SVNRepository.php');
require_once('plugins/scm/lib/ScmProjectStats.php');
require_once('plugins/scm/lib/ScmUserStats.php');

// this could be a lengthy process
set_time_limit(0);

/**
 * Cronjob that parses the SVN history
 */
class ParseSVNHistoryCronjob extends Cronjob {
	private $svnbin = '';
	
	public function __construct() {
		$this->jobname = basename(__FILE__);
		parent::__construct();
	}	
	
	public function run() {
		// find the SVN binary
		$this->svnbin = SystemUtils::buildFilePath(GFConf::get('scmsvn.svnpath'), '/svn');
		if (!file_exists($this->svnbin)) {
			$this->logCron('SVN binary '.$this->svnbin.' not found');
			$this->commitLog();
			exit(1);
		}
		
		
		// lets get all projects that use the plugin
		$c = new Criteria();
		$c->add(PluginPeer::PLUGIN_NAME, 'scmsvn');
		$project_plugins = ProjectPluginPeer::doSelectJoinPlugin($c);
		
		foreach ($project_plugins as $project_plugin) {
			$project = $project_plugin->getProject(); // get the project
			$repos = new SVNRepository($project);
			
			// repository hasn't been created yet... there isn't a history file
			if (!$repos->isCreated()) continue;
			
			$this->parseHistory($repos, $project);
		}
	}
	
	public function parseHistory(SVNRepository $repos, Project $project) {
		$root = $repos->getRoot();

		$output = array();

		$day_parameter = $this->getCronjobParameter(0);
		if (strtolower($day_parameter) == 'all') {
			// Parsing all days -- retrieve the full log
			$cmd = $this->svnbin.' log file://'.$root.' -q -v --xml';
		} else {
			$beg=(time()-(4*86400));
			$end=time();
			$cmd = $this->svnbin.' log file://'.$root.' -r {'.date('Ymd',$beg).'}:{'.date('Ymd',$end).'} -q -v --xml';
		}

		$res = SystemUtils::execute($cmd, $output);
		if ($res != 0) {		// error in command
			return;
		}
		
		$xml = implode("\n", $output);
		$days_data = $this->parseHistoryXML($xml);
		
		foreach ($days_data as $day => $day_data) {
			$c = new Criteria();
			$c->add(ScmProjectStatsPeer::STAT_DATE, $day);
			$c->add(ScmProjectStatsPeer::PROJECT_ID, $project->getProjectId());
			$stats = ScmProjectStatsPeer::doSelectOne($c);
			if (is_null($stats)) {
				$stats = new ScmProjectStats();
				$stats->setProjectId($project->getProjectId());
				$stats->setStatDate($day);
			}
			
			$stats->setCommits($day_data['commits']);
			$stats->setAdds($day_data['adds']);
			$stats->setCheckouts(0);		// No checkouts information for SVN
			$stats->save();
			
			// now store the user information
			foreach ($day_data['user_data'] as $user_name => $user_data) {
				$user = UserPeer::retrieveByUnixName($user_name);
				if (is_null($user)) {
					// Not a GForge user -- use the undefined user
					$user = UserPeer::retrieveByPk(User::UNDEFINED_USER_ID);
					if (is_null($user)) continue;
				}
				
				$c = new Criteria();
				$c->add(ScmUserStatsPeer::STAT_DATE, $day);
				$c->add(ScmUserStatsPeer::PROJECT_ID, $project->getProjectId());
				$c->add(ScmUserStatsPeer::USER_ID, $user->getUserId());
				
				$stats = ScmUserStatsPeer::doSelectOne($c);
				if (!$stats) {
					$stats = new ScmUserStats();
					$stats->setStatDate($day);
					$stats->setProjectId($project->getProjectId());
					$stats->setUserId($user->getUserId());
				}
				
				$stats->setCommits($user_data['commits']);
				$stats->setAdds($user_data['adds']);
				$stats->setCheckouts(0);
				$stats->save();
			}
		}
	}
	
	public function parseHistoryXML($xml) {
		$days = array();


		// See which day(s) we want to parse
		$day_parameter = $this->getCronjobParameter(0);
		if (strtolower($day_parameter) == 'all') {
			// Parse all days?
			$parse_all = true;
		} else {
			$parse_all = false;
			
			$days_to_parse = array();			
			
			if ($day_parameter && ($time = strtotime($day_parameter)) !== -1) {
				$days_to_parse[] = date('Y-m-d', $time);
			} else {
				// Parse today and yesterday. This script is usually ran once per day. Parsing
				// the two days will ensure yesterday's information is correct, otherwise
				// the information between this script's last run and yesterday midnight
				// would get lost.
				$today = time();
				$yesterday = time()-(60*60*24);
				
				$days_to_parse[] = date('Y-m-d', $today);
				$days_to_parse[] = date('Y-m-d', $yesterday);
			}
		}

		$xml = simplexml_load_string($xml);
		foreach ($xml->logentry as $entry) {
			$author = (string)$entry->author;

			$day = substr($entry->date, 0, 10);
			
			if (!$parse_all && !in_array($day, $days_to_parse)) {
				// we don't want to parse this day
				continue;
			}
			
			if (!array_key_exists($day, $days)) {
				$days[$day] = array();
				$days[$day]['adds'] = 0;
				$days[$day]['commits'] = 0;
				$days[$day]['user_data'] = array();
			}
			
			if (!array_key_exists($author, $days[$day]['user_data'])) {
				$days[$day]['user_data'][$author] = array();
				$days[$day]['user_data'][$author]['adds'] = 0;
				$days[$day]['user_data'][$author]['commits'] = 0;
			}
			
			foreach ($entry->paths->path as $path) {
				if ($path['action'] == 'A') {
					$days[$day]['adds']++;
					$days[$day]['user_data'][$author]['adds']++;
				}
				if ($path['action'] == 'M') {
					$days[$day]['commits']++;
					$days[$day]['user_data'][$author]['commits']++;
				}
			}
		}
		
		return $days;
	}
}

$cronjob = new ParseSVNHistoryCronjob();
$cronjob->run();
?>
