<?php
/**
 * Copyright (c) 2005-2008 GForge, LLC
 *
 * @version   $Id: SVNRepository.php 4415 2010-03-10 16:15:09Z tperdue $
 * @author Marcelo Mottalli	marcelo@gforgegroup.com
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once('SystemUtils.php');
require_once('plugins/scmsvn/lib/exceptions/SVNException.php');
require_once('plugins/scm/lib/SCMRepository.php');
require_once('plugins/scm/lib/ScmCommitFilter.php');
require_once('plugins/scm/lib/ScmCommitFilterProject.php');


class SVNRepository extends SCMRepository {
	private $repositoryPath;
	private $svnRoot;
	private $useSSH = false;
	private $revisionCache = array();
	
	public function __construct(Project $project) {
		parent::__construct($project);

		$this->svnRoot = GFConf::get('scmsvn.localSVNRoot');
		$this->repositoryPath = SystemUtils::buildFilePath($this->svnRoot, '/'.$this->repositoryName);
		$this->svnpath = GFConf::get('scmsvn.svnpath');
		$this->useSSH = GFConf::get('scmsvn.useSSH');
	}
	
	public function getType() {
		return 'svn';
	}
	
	/**
	 * @return	bool	True if the repository is created
	 */
	public function isCreated() {
		return is_dir($this->repositoryPath);
	}
	
	/**
	 * Create and initialize this repository
	 * @return	bool	True on success
	 */
	public function create() {
		if ($this->isCreated()) {
			throw new SVNException('Tried to create an existing repository');
		}
		
		if ($this->useSSH) {
			$output = array();
			$unix_group_name = $this->project->getUnixName();
			if (!SystemUtils::groupExists($unix_group_name)) {
				throw new SVNException('Tried to create repository for project '.$unix_group_name.' but UNIX group doesn\'t exist');
			}
	
			$repos_type = GFConf::get('scmsvn.reposType');
			$file_owner = 'nobody:'.$unix_group_name; // what do we need this for?

			mkdir($this->repositoryPath, 0770, true);
	 		$cmd = $this->svnpath.'/svnadmin create --fs-type '.$repos_type.' '.$this->repositoryPath;
	
	 		if (SystemUtils::execute($cmd.' init', $output) != 0) {
	 			return false;
	 		}
	 		$this->createDefaultSkeleton();
	 		$this->updateRepository();
		} else {

			mkdir($this->repositoryPath, 02770, true);
			
			$repos_type = GFConf::get('scmsvn.reposType');
			$cmd = $this->svnpath.'/svnadmin create --fs-type '.$repos_type.' '.$this->repositoryPath;
			
			$output = array();
			if (SystemUtils::execute($cmd.' init', $output) != 0) {
				return false;
			}
	
			// Change the repository permissions to apache user & group
			$sys_apache_user = GFConf::get('system.apacheUser');
			$sys_apache_group = GFConf::get('system.apacheGroup');
			$file_owner = $sys_apache_user.':'.$sys_apache_group;
			SystemUtils::execute('chown -R '.$file_owner.' '.$this->repositoryPath, $output);
			$this->createDefaultSkeleton();
			$this->updateRepository();
		}
		return true;
	}
	
	/**
	 * Change the repository's permissions in order to enable anonymous access
	 */
	public function enableAnonAccess() {
		SystemUtils::execute('chmod -R o+r '.$this->repositoryPath, $output);
		// Anonymous can't write
		SystemUtils::execute('chmod -R o-w '.$this->repositoryPath, $output);
		SystemUtils::execute('chmod o+x '.$this->repositoryPath, $output);
		// Allow access to directories
		SystemUtils::execute('find '.$this->repositoryPath.' -type d -exec chmod o+x {} \;', $output);
	}
	
	/**
	 * Change the repository's permissions in order to disable anonymous access
	 */
	public function disableAnonAccess() {
		SystemUtils::execute('chmod -R o-r '.$this->repositoryPath, $output);
		SystemUtils::execute('chmod -R o-w '.$this->repositoryPath, $output);
		SystemUtils::execute('chmod o-x '.$this->repositoryPath, $output);
		// Disable access to directories
		SystemUtils::execute('find '.$this->repositoryPath.' -type d -exec chmod o-x {} \;', $output);
	}
	

	public function updateRepository() {
		$preCommitFile = SystemUtils::buildFilePath($this->repositoryPath, '/hooks/pre-commit');
		$postCommitFile = SystemUtils::buildFilePath($this->repositoryPath, '/hooks/post-commit');
		
		if (!file_exists($preCommitFile)) {
			$preCommit = $this->getDefaultPreCommitFile();
		} else {
			$preCommit = file($preCommitFile);
			$preCommit = array_map("trim", $preCommit);
		}
		
		if (!file_exists($postCommitFile)) {
			$postCommit = $this->getDefaultPostCommitFile();
		} else {
			$postCommit = file($postCommitFile);
			$postCommit = array_map("trim", $postCommit);
		}
		
		// Add misc configuration directives
		$this->updatePreCommitHooks($preCommit);
		$this->updatePostCommitHooks($postCommit);
		$this->updateSVNMail($postCommit);

		$output = array();

		if ($this->useSSH) {
			// Update the permissions
			$unix_group_name = $this->project->getUnixName();
			
			// This can cause problems if there's an ongoing transaction, see [#4131]
			//SystemUtils::execute('chown -R nobody:'.$unix_group_name.' '.$this->repositoryPath, $output);	
			// Safe change
			SystemUtils::execute('chgrp -R '.$unix_group_name.' '.$this->repositoryPath, $output);
			
			
			// User and group should have read+write access
			// (See http://svnbook.red-bean.com/en/1.0/ch06s05.html, 'The svn+ssh server checklist')
			SystemUtils::execute('chmod -R u+r '.$this->repositoryPath, $output);
			SystemUtils::execute('chmod -R u+w '.$this->repositoryPath, $output);
			SystemUtils::execute('chmod -R g+r '.$this->repositoryPath, $output);
			SystemUtils::execute('chmod -R g+w '.$this->repositoryPath, $output);
			SystemUtils::execute('chmod u+x '.$this->repositoryPath, $output);
			SystemUtils::execute('chmod g+x '.$this->repositoryPath, $output);

			$project_scm = $this->getScmConfiguration();

			if ($this->project->getIsPublic() && $project_scm->getEnableAnonRead()) {
				$this->enableAnonAccess();
			} else {
				// Private project or anonymous SVN disabled
				$this->disableAnonAccess();
			}

	 		$this->writeFile($preCommitFile, implode("\n", $preCommit));
	 		SystemUtils::execute('chmod +x '.$preCommitFile, $output);
			SystemUtils::execute('chmod 770 '.$preCommitFile, $output);
			chown($preCommitFile, 'nobody');
			chgrp($preCommitFile, $unix_group_name);

	 		$this->writeFile($postCommitFile, implode("\n", $postCommit));
	 		SystemUtils::execute('chmod +x '.$postCommitFile, $output);
			SystemUtils::execute('chmod 770 '.$postCommitFile, $output);
			chown($postCommitFile, 'nobody');
			chgrp($postCommitFile, $unix_group_name);

		} else {
			$sys_apache_user = GFConf::get('system.apacheUser');
			$sys_apache_group = GFConf::get('system.apacheGroup');
			
			$this->writeFile($preCommitFile, implode("\n", $preCommit));
			SystemUtils::execute('chmod +x '.$preCommitFile, $output);
			SystemUtils::execute('chmod 700 '.$preCommitFile, $output);

			$this->writeFile($postCommitFile, implode("\n", $postCommit));
			SystemUtils::execute('chmod +x '.$postCommitFile, $output);
			SystemUtils::execute('chmod 700 '.$postCommitFile, $output);

			$file_owner = $sys_apache_user.':'.$sys_apache_group;
			SystemUtils::execute('chown -R '.$file_owner.' '.$this->repositoryPath, $output);			
		}
	}
	
	public function updateSVNMail(&$lines) {
		// Read the preferences
		$preferences = $this->getScmPreferences();

 		if (!is_null($preferences) && $preferences->getMailCommits()) {
			$this->addSVNMail($lines);
		} else {
			$this->removeSVNMail($lines);
		}
	}

	
	/**
	* This function creates the post-commit file in svn hooks (if it doesn't exist)
	*/
	function getDefaultPostCommitFile() {
		return array(
			'#!/bin/sh',
			'REPOS="$1"',
			'REV="$2"',
			''	// empty line
		);
	}
	
	/**
	* This function creates the pre-commit file in svn hooks (if it doesn't exist)
	*/
	function getDefaultPreCommitFile() {
		return array(
			'#!/bin/sh',
			'REPOS="$1"',
			'TXN="$2"',
			''	// empty line
		);
	}

	public function addSVNMail(&$lines) {
		// First check if we have mailman configured
		if (!$this->project->usesPlugin('mailman')) {
			return;
		}

		$found = false;
		foreach ($lines as $line) {
			if (strpos($line, 'commit-email.pl') !== false) {
				$found = true;
				break;
			}
		}
		// File already has SVN mail
		if ($found) {
			return;
		}
		
		$commitScript = SystemUtils::buildFilePath(GFConf::get('system.pluginsPath'), '/scmsvn/bin/commit-email.pl');
		$lines[] = $commitScript.' "$REPOS" "$REV" '.$this->repositoryName.'-commits@'.GFConf::get('mailman.listsHost');
	}
	
	public function removeSVNMail(&$lines) {
		$new_content = array();
		for ($i=0; $i < count($lines); $i++) {
			$line = trim($lines[$i]);
			
			if (strpos($line, '/scmsvn/bin/commit-email.pl') !== false) {
				// Found commit-email line... skip current line
			} else {
				$new_content[] = $line;
			}
		}
		$lines = $new_content;
	}
	
	/**
	 * updatePostCommitHooks - Add the commit hooks for ActionManager
	 */
	public function updatePostCommitHooks(&$lines) {
		$found = false;
		foreach ($lines as $line) {
			if (strpos($line, '#BEGIN GForge commit hooks') !== false) {
				$found = true;
				break;
			}
		}
		// File already has commit hooks
		if ($found) {
			return;
		}

		$postscript = SystemUtils::buildFilePath(GFConf::get('system.pluginsPath'), '/scmsvn/bin/post_commit.php');
        
        $lines[] = '#BEGIN GForge commit hooks';
        $lines[] = 'php -q -d include_path='.ini_get('include_path').' '.$postscript.' "$REPOS" "$REV"';
        $lines[] = '#END GForge commit hooks';
	}
	
	/**
	 * updatePreCommitHooks - Add pre-commit hooks
	 */
	public function updatePreCommitHooks(&$lines) {
		$this->insertCommitFilters($lines);

		$found = false;
		foreach ($lines as $line) {
			if (strpos($line, '#BEGIN GForge pre-commit checks') !== false) {
				$found = true;
				break;
			}
		}
		// File already has commit hooks
		if ($found) {
			return;
		}

		$precommitcheckscript = SystemUtils::buildFilePath(GFConf::get('system.pluginsPath'), '/scmsvn/bin/pre_commit_checks.php');
        
        $lines[] = '#BEGIN GForge pre-commit checks';
        $lines[] = 'php -q -d include_path='.ini_get('include_path').' '.$precommitcheckscript.' "$REPOS" "$TXN"';
        $lines[] = '#END GForge pre-commit checks';
	}
	
	/**
	 * Inserts the commit filter commands
	 */
	private function insertCommitFilters(&$lines_src) {
		$lines_dest = array();

		for ($i = 0; $i < count($lines_src); $i++) {
			$line = $lines_src[$i];
			if (strpos($line, '#BEGIN GForge filters') !== false) {
				// Seek to the end of the marker
				for (; $i < count($lines_src); $i++) {
					$line = $lines_src[$i];
					if (strpos($line, '#END GForge filters') !== false) {
						break;
					}
				}
			} else {
				$lines_dest[] = $line;
			}
		}
		
		// Now append the filters

		$c = new Criteria();
		$c->add(ProjectPluginPeer::PROJECT_ID, $this->project->getProjectId());
		$c->addJoin(ProjectPluginPeer::PLUGIN_ID, PluginPeer::PLUGIN_ID);
		$c->add(PluginPeer::PLUGIN_NAME, ScmsvnPlugin::getName());		// Make sure we're not including filters we shouldn't
		$c->addJoin(ScmCommitFilterProjectPeer::PROJECT_ID, ProjectPluginPeer::PROJECT_ID);
		$c->addJoin(ScmCommitFilterPeer::PLUGIN_ID, PluginPeer::PLUGIN_ID);
		$c->addJoin(ScmCommitFilterProjectPeer::SCM_COMMIT_FILTER_ID, ScmCommitFilterPeer::SCM_COMMIT_FILTER_ID);
		$filters = ScmCommitFilterPeer::doSelect($c);
		
		$lines_dest[] = '#BEGIN GForge filters';
		foreach ($filters as $filter) {
			$command = $filter->getFilterString();
			$command = StringUtils::unHtmlSpecialChars($command);
			$command = preg_replace('/\\$PROJECT/i', $this->project->getUnixName(), $command);
			$command = preg_replace('/\\$PATH/i', '\\$1', $command);
			$command = preg_replace('/\\$FILE/i', '\\$2', $command);
			$lines_dest[] = $command;
		}
		$lines_dest[] = '#END GForge filters';
		
		$lines_src = $lines_dest;
	}
	/**
	 * checkoutFile - Retrieve a file into a temporary directory from the repository
	 *
	 * @param	string $filename	File to checkout
	 * return	string	The name of the copy where the file has been saved
	 */

	public function checkoutFile($filename) {
	}
	
	/**
	 * commitFile - Commit a file to the repository
	 *
	 * @param	string $file	File to commit
	 * @param	string $message	Commit message
	 */
	function commitFile($file,$message="Automatic updated by scmsvn") {
	}
	
	/**
	 * releaseFile - Remove the file that was just checked out from the repository
	 * @see checkoutFile
	 */
	function releaseFile($file) {
		// $file is something like /tmp/(tmp_dir)/path/to/file
		// we must delete /tmp/tmp_dir
		if (!preg_match("/^(\/tmp\/[^\/]+)\/.*/", $file, $result)) {		// Make sure the dir is under /tmp
			$this->err = "Trying to release a directory not in /tmp. Skipping...";
			$this->logCron();
			return;
		}
		$dir = $result[1];
		
		// this shouldn't happen... but add it as a security check
		if (!preg_match('/[^\\/]/',$dir)) {
			$this->err = "Trying to delete root dir. Skipping...";
			$this->logCron();
			return;
		}
		$dir = escapeshellarg($dir);
		system("rm -rf ".$dir);
	}

	/**
	 * Writes to a file with locking
	 * @param	string $filePath
	 * @param	mixed $content	Contents of the file. Can be a string or an array with the lines
	 * @param	int	0 write to the file, 1 append to the file, 2 truncate, rewind and rewrite
	 */
	private function writeFile($filePath, $content, $append=0) {
		if (is_array($content)) {
			$content = implode("\n", $content);
		}
		if ($append==2) {
			$file = fopen($filePath, 'a');
			flock($file, LOCK_EX);
			ftruncate($file, 0);
			rewind($file);
			if(!empty($content)) {
				fwrite($file, $content);
			}
			flock($file, LOCK_UN);
			fclose($file);	
		} else {
			if ($append == 1) {
				$file = fopen($filePath, 'a');
			} else {
				$file = fopen($filePath, 'w');
			}
			flock($file, LOCK_EX);
			if(!empty($content)) {
				fwrite($file, $content);
			}
			flock($file, LOCK_UN);
			fclose($file);
		}
	}
	
	/**
	 * Returns the SVNROOT for this repository
	 * @return	string
	 */
	public function getRoot() {
		return $this->repositoryPath;
	}
	
	/**
	 * Returns the for ALL repositories
	 * @return	string
	 */
	public function getScmRoot() {
		return GFConf::get('scmsvn.localSVNRoot');
	}
	
	/**
	 * Create the SVN repo default skeleton
	 * /trunk, /branches and /tags
	 *
	 */
	public function createDefaultSkeleton() {
		$path = "file://{$this->repositoryPath}/trunk file://{$this->repositoryPath}/branches file://{$this->repositoryPath}/tags";
		$cmd = $this->svnpath."/svn mkdir $path -m 'Initial Repository Skeleton'";
		SystemUtils::execute($cmd, $output);
	}
	
	public function doLocalCheckout($destdir, $module=null) {
		$path = $this->repositoryPath;
		if ($module) {
			$path = SystemUtils::buildFilePath($path, $module);			// Will be something like /svnroot/project/trunk/module
		}
		
		$cmd = SystemUtils::buildFilePath($this->svnpath, 'svn')." co file://$path $destdir";
		$output = array();
		$error = SystemUtils::execute($cmd, $output);
		
		if ($error) {
			throw new Exception("Unable to checkout the code - Error code $error - Command was '$cmd'");
		}
	}

	public function getCheckedoutFileRevision($checkout_dir, $file) {
		if (!isset($this->revisionCache[$checkout_dir])) {
			$this->revisionCache[$checkout_dir] = array();
		} elseif (isset($this->revisionCache[$checkout_dir][$file])) {
			return $this->revisionCache[$checkout_dir][$file];
		}

		$fullpath = SystemUtils::buildFilePath($checkout_dir, $file);
		
		$output = array();
		$cmd = $this->svnpath."/svn info --xml $fullpath";
		$error = SystemUtils::execute($cmd, $output);		// Note: DON'T use $explode_output=false on the call to SystemUtils because we need the return value
		
		if ($error) {
			$rev = -1;
		} else {
			$output = implode("\n", $output);
			$xml = new SimpleXMLElement($output);
			if (!$xml || !isset($xml->entry)) {
				$rev = -1;
			} else {
				$rev = intval($xml->entry['revision']);
			}
		}
		
		// Cache for speeded-up access (can be called several times from the scm_index cronjob)
		$this->revisionCache[$checkout_dir][$file] = $rev;
		return $rev;
	}
}
?>
